/*
 * arch/ppc/boot/simple/mv64360_tty.c
 * 
 * Bootloader version of the embedded MPSC/UART driver for the MV64360.
 * Note: Due to 64260A errata, DMA will be used for UART input (via SDMA).
 *
 * Author: Mark A. Greer <mgreer@mvista.com>
 *
 * Copyright 2001 MontaVista Software Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

/* This code assumes that the data cache has been disabled (L1, L2, L3). */

#include <linux/config.h>
#include <linux/serialP.h>
#include <linux/serial_reg.h>
#include <asm/serial.h>
#include <asm/mv64360_defs.h>

extern void udelay(long);
static void stop_dma(int chan);

#ifdef	CONFIG_MV64360_NEW_BASE
static u32	mv64360_base = CONFIG_MV64360_NEW_REG_BASE;
#else
static u32	mv64360_base = CONFIG_MV64360_ORIG_REG_BASE;
#endif

inline unsigned
mv64360_in_le32(volatile unsigned *addr)
{
	unsigned ret;

	__asm__ __volatile__("lwbrx %0,0,%1; eieio" : "=r" (ret) :
				     "r" (addr), "m" (*addr));
	return ret;
}

inline void
mv64360_out_le32(volatile unsigned *addr, int val)
{
	__asm__ __volatile__("stwbrx %1,0,%2; eieio" : "=m" (*addr) :
				     "r" (val), "r" (addr));
}

#define MV64360_REG_READ(offs)						\
	(mv64360_in_le32((volatile uint *)(mv64360_base + (offs))))
#define MV64360_REG_WRITE(offs, d)					\
        (mv64360_out_le32((volatile uint *)(mv64360_base + (offs)), (int)(d)))


static struct {
	u32	sdc;
	u32	sdcm;
	u32	rx_desc;
	u32	rx_buf_ptr;
	u32	scrdp;
	u32	tx_desc;
	u32	sctdp;
	u32	sftdp;
} sdma_regs;

#define	SDMA_REGS_INIT(channel) {						           \
	sdma_regs.sdc        = MV64360_SDMA_CONFIG_REG(channel);		   \
	sdma_regs.sdcm       = MV64360_SDMA_COMMAND_REG(channel);	       	   \
	sdma_regs.scrdp      = MV64360_SDMA_CURRENT_RX_DESCRIPTOR_POINTER(channel);\
	sdma_regs.sctdp      = MV64360_SDMA_CURRENT_TX_DESCRIPTOR_POINTER(channel);\
	sdma_regs.sftdp      = MV64360_SDMA_FIRST_TX_DESCRIPTOR_POINTER(channel);  \
}

typedef struct {
	volatile u16 bufsize;
	volatile u16 bytecnt;
	volatile u32 cmd_stat;
	volatile u32 next_desc_ptr;
	volatile u32 buffer;
} mv64360_rx_desc_t;

typedef struct {
	volatile u16 bytecnt;
	volatile u16 shadow;
	volatile u32 cmd_stat;
	volatile u32 next_desc_ptr;
	volatile u32 buffer;
} mv64360_tx_desc_t;

#define	MAX_RESET_WAIT	10000
#define	MAX_TX_WAIT	10000

#define	RX_NUM_DESC	2
#define	TX_NUM_DESC	2

#define	RX_BUF_SIZE	16
#define	TX_BUF_SIZE	16

static mv64360_rx_desc_t rd[RX_NUM_DESC] __attribute__ ((aligned(32)));
static mv64360_tx_desc_t td[TX_NUM_DESC] __attribute__ ((aligned(32)));

static char rx_buf[RX_NUM_DESC * RX_BUF_SIZE] __attribute__ ((aligned(32)));
static char tx_buf[TX_NUM_DESC * TX_BUF_SIZE] __attribute__ ((aligned(32)));

static int cur_rd = 0;
static int cur_td = 0;


#define	RX_INIT_RDP(rdp) {						\
	(rdp)->bufsize = 2;						\
	(rdp)->bytecnt = 0;						\
	(rdp)->cmd_stat = MV64360_SDMA_DESC_CMDSTAT_L |			\
			  MV64360_SDMA_DESC_CMDSTAT_F |			\
			  MV64360_SDMA_DESC_CMDSTAT_O;			\
}

unsigned long
serial_init(int chan, void *ignored)
{
	u32     cdv;
	int	i;

	stop_dma(0);
	stop_dma(1);

	if (chan != 1) {
		chan = 0;  /* default to chan 0 if anything but 1 */
		SDMA_REGS_INIT(0);
	}
	else {
		SDMA_REGS_INIT(1);
	}

	/* Set up ring buffers */
	for (i=0; i<RX_NUM_DESC; i++) {
		RX_INIT_RDP(&rd[i]);
		rd[i].buffer = (u32)&rx_buf[i * RX_BUF_SIZE];
		rd[i].next_desc_ptr = (u32)&rd[i+1];
	}
	rd[RX_NUM_DESC - 1].next_desc_ptr = (u32)&rd[0];

	for (i=0; i<TX_NUM_DESC; i++) {
		td[i].bytecnt = 0;
		td[i].shadow = 0;
		td[i].buffer = (u32)&tx_buf[i * TX_BUF_SIZE];
		td[i].cmd_stat = MV64360_SDMA_DESC_CMDSTAT_F |
				 MV64360_SDMA_DESC_CMDSTAT_L;
		td[i].next_desc_ptr = (u32)&td[i+1];
	}
	td[TX_NUM_DESC - 1].next_desc_ptr = (u32)&td[0];

	/* Set MPSC Routing */
        MV64360_REG_WRITE( MV64360_MPSC_ROUTING_REG, 0x3ffffe38);

	/* MPSC 0/1 Rx & Tx get clocks BRG0/1 */
        MV64360_REG_WRITE(MV64360_MPSC_RX_CLOCK_ROUTING_REG, 0x00000100);
        MV64360_REG_WRITE(MV64360_MPSC_TX_CLOCK_ROUTING_REG, 0x00000100);

	/* clear pending interrupts */
        MV64360_REG_WRITE(MV64360_SDMA_MASK_REG, 0);

	MV64360_REG_WRITE(MV64360_SDMA_CURRENT_RX_DESCRIPTOR_POINTER(chan), &rd[0]);
	MV64360_REG_WRITE(MV64360_SDMA_CURRENT_TX_DESCRIPTOR_POINTER(chan),
		&td[TX_NUM_DESC - 1]);
	MV64360_REG_WRITE( MV64360_SDMA_FIRST_TX_DESCRIPTOR_POINTER(chan),
		&td[TX_NUM_DESC - 1]);

	MV64360_REG_WRITE(MV64360_SDMA_CONFIG_REG(chan),
			  MV64360_SDMA_SDC_RFT  | MV64360_SDMA_SDC_SFM  |
			  MV64360_SDMA_SDC_BLMR | MV64360_SDMA_SDC_BLMT |
			  (3 << 12));

	cdv = ((CONFIG_MV64360_BRG_CLK_RATE/(32*CONFIG_SERIAL_CONSOLE_BAUD))-1);
	MV64360_REG_WRITE(MV64360_BRG_CONFIG_REG(chan),
		((CONFIG_MV64360_CLKSRC << 18) | (1 << 16) | cdv));

	/* Put MPSC into UART mode, no null modem, 16x clock mode */
	MV64360_REG_WRITE(MV64360_MPSC_MAIN_CONFIG_LOW(chan), 0x000004c4);
	MV64360_REG_WRITE(MV64360_MPSC_MAIN_CONFIG_HIGH(chan), 0x04400400);

        MV64360_REG_WRITE(MV64360_MPSC_CHANNEL_REG1(chan), 0);
        MV64360_REG_WRITE(MV64360_MPSC_CHANNEL_REG9(chan), 0);
        MV64360_REG_WRITE(MV64360_MPSC_CHANNEL_REG10(chan), 0);
        MV64360_REG_WRITE(MV64360_MPSC_CHANNEL_REG3(chan), 4);
        MV64360_REG_WRITE(MV64360_MPSC_CHANNEL_REG4(chan), 0);
        MV64360_REG_WRITE(MV64360_MPSC_CHANNEL_REG5(chan), 0);
        MV64360_REG_WRITE(MV64360_MPSC_CHANNEL_REG6(chan), 0);
        MV64360_REG_WRITE(MV64360_MPSC_CHANNEL_REG7(chan), 0);
        MV64360_REG_WRITE(MV64360_MPSC_CHANNEL_REG8(chan), 0);

	/* 8 data bits, 1 stop bit */
	MV64360_REG_WRITE(MV64360_MPSC_PROTOCOL_CONFIG(chan), (3 << 12));

	MV64360_REG_WRITE(MV64360_SDMA_COMMAND_REG(chan),
		(1<<7));

	MV64360_REG_WRITE(MV64360_MPSC_CHANNEL_REG1(chan),
		(1<<31));

	udelay(100);

	return (ulong)chan;
}

static void
stop_dma(int chan)
{
	int	i;

	/* Abort SDMA Rx, Tx */
	MV64360_REG_WRITE(MV64360_SDMA_COMMAND_REG(chan),
		MV64360_SDMA_SDCM_AR | MV64360_SDMA_SDCM_STD);

	for (i=0; i<MAX_RESET_WAIT; i++) {

		if ( (MV64360_REG_READ(MV64360_SDMA_COMMAND_REG(chan)) & 
		      (MV64360_SDMA_SDCM_AR | MV64360_SDMA_SDCM_STD)) == 0)
			break;
		udelay(100);
	}
	
	return;
}

static int
wait_for_ownership(void)
{
	int	i;

	for (i=0; i<MAX_TX_WAIT; i++) {
		if ((MV64360_REG_READ(sdma_regs.sdcm) &
					(1 << 23)) == 0) break;
		udelay(1000);
	}

	return (i < MAX_TX_WAIT);
}

void
serial_putc(unsigned long com_port, unsigned char c)
{
	mv64360_tx_desc_t	*tdp;

	if (wait_for_ownership() == 0) return;

	tdp = &td[cur_td];
	if (++cur_td >= TX_NUM_DESC) cur_td = 0;

	*(unchar *)(tdp->buffer ^ 7) = c;
	tdp->bytecnt = 1;
	tdp->shadow = 1;
	tdp->cmd_stat = MV64360_SDMA_DESC_CMDSTAT_L |
		MV64360_SDMA_DESC_CMDSTAT_F | MV64360_SDMA_DESC_CMDSTAT_O;

	MV64360_REG_WRITE(sdma_regs.sctdp, tdp);
	MV64360_REG_WRITE(sdma_regs.sftdp, tdp);
	MV64360_REG_WRITE(sdma_regs.sdcm,
			  MV64360_REG_READ(sdma_regs.sdcm) | MV64360_SDMA_SDCM_TXD);

	return;
}

unsigned char
serial_getc(unsigned long com_port)
{
	mv64360_rx_desc_t	*rdp;
	unchar			c = '\0';

	rdp = &rd[cur_rd];

	if ((rdp->cmd_stat & (MV64360_SDMA_DESC_CMDSTAT_O |
			      MV64360_SDMA_DESC_CMDSTAT_ES)) == 0) {
		c = *(unchar *)(rdp->buffer ^ 7);
		RX_INIT_RDP(rdp);
		if (++cur_rd >= RX_NUM_DESC) cur_rd = 0;
	}

	return c;
}

int
serial_tstc(unsigned long com_port)
{
	mv64360_rx_desc_t	*rdp;
	int			loop_count = 0;
	int			rc = 0;

	rdp = &rd[cur_rd];

	/* Go thru rcv desc's until empty looking for one with data (no error)*/
	while (((rdp->cmd_stat & MV64360_SDMA_DESC_CMDSTAT_O) == 0) &&
	       (loop_count++ < RX_NUM_DESC)) {

		/* If there was an error, reinit the desc & continue */
		if ((rdp->cmd_stat & MV64360_SDMA_DESC_CMDSTAT_ES) != 0) {
			RX_INIT_RDP(rdp);
			if (++cur_rd >= RX_NUM_DESC) cur_rd = 0;
			rdp = (mv64360_rx_desc_t *)rdp->next_desc_ptr;
		}
		else {
			rc = 1;
			break;
		}
	}

	return rc;
}

void
serial_close(unsigned long com_port)
{
	stop_dma(com_port);
	return;
}
