/*
 *
 *  85xx_dma_priv.h, version 1.0 
 *  Private info for general purpose DMA controller
 *
 *  This software may be used and distributed according to the terms of
 *  the GNU Public License, Version 2, incorporated herein by reference.
 *
 *  Contact:  <kumar.gala@motorola.com>
 *
 */

#ifndef __MPC85XX_DMA_PRIV__
#define __MPC85XX_DMA_PRIV__

#include <linux/module.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/tqueue.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/stddef.h>
#include <linux/string.h>
#include <linux/types.h>
#include <asm/bitops.h>
#include <asm/io.h>
#include <linux/interrupt.h>
#include <asm/mpc85xx.h>

#include "85xx_dma.h"

#define NDEBUG

#ifndef NDEBUG
#define Debug(message)  printk message
#define assert(expr)    \
    if (!(expr))        \
        printk("Assertion failure:  (%s) in %s, %s:%d\n", \
                #expr, __FUNCTION__, __FILE__, __LINE__); else
#else				/* NDEBUG */
#define Debug(message)
#define assert(expr)
#endif				/* NDEBUG */

#define MIN_POOL_SIZE	10
#define MAX_POOL_SIZE	40

#define DMA_ALIGN		0x20
#define DMA_NUM_CHANNELS	4

#define DEFAULT_IRQ		MPC85xx_IRQ_DMA0

#define PRINT_PREFIX    "85xx_dma: "

#define	DMA_MR_EOSIE			(1<<9)	/* End-of-Segment Interrupt Enable */
#define	DMA_MR_EOLNIE			(1<<8)	/* End-of-Links Interrupt Enable */
#define	DMA_MR_EOLSIE			(1<<7)	/* End-of-Lists Interrupt Enable */
#define DMA_MR_PEIE			(1<<6)	/* Program Error Interrupt Enable */
#define DMA_MR_XFE			(1<<5)	/* Extended Features Enable (striding) */
#define DMA_MR_CS			(1<<0)	/* Channel Start */

#define DMA_MODE_PRC_READ		(0<<10)
#define DMA_MODE_PRC_READ_LINE		(1<<10)
#define DMA_MODE_PRC_READ_MULT		(2<<10)
#define DMA_MODE_EIE			(1<<8)
#define DMA_MODE_EOTIE			(1<<7)
#define DMA_MODE_DIRECT			(1<<2)
#define DMA_MODE_START			(1<<0)

#define DMA_SR_TE 	(1<<7)
#define DMA_SR_CH 	(1<<5)
#define DMA_SR_PEI	(1<<4)
#define DMA_SR_EOLNI	(1<<3)
#define DMA_SR_CB	(1<<2)
#define DMA_SR_EOSI	(1<<1)
#define DMA_SR_EOLSI	(1<<0)

#define DMA_ERROR	(DMA_SR_TE | DMA_SR_PEI)
#define DMA_SR_ALL	(DMA_SR_TE | DMA_SR_CH | \
      			 DMA_SR_PEI | DMA_SR_EOLNI | \
			 DMA_SR_CB | DMA_SR_EOSI | DMA_SR_EOLSI)

#define DMA_SATR_SSME				(0x1<<24)
#define DMA_SATR_SREADTTYPE_NOSNOOP		(0x4<<16)
#define DMA_SATR_SREADTTYPE_SNOOP		(0x5<<16)
#define DMA_SATR_SREADTTYPE_UNLOCK_L2		(0x7<<16)

#define DMA_DATR_DSME				(0x1<<24)
#define DMA_DATR_DWRITETTYPE_NOSNOOP		(0x4<<16)
#define DMA_DATR_DWRITETTYPE_SNOOP		(0x5<<16)
#define DMA_DATR_DWRITETTYPE_ALLOC_L2		(0x6<<16)
#define DMA_DATR_DWRITETTYPE_ALLOC_LOCK_L2	(0x7<<16)

#define DMA_CLNDAR_GLEN		(1<<4)

#define DMA_LINK_DESC_EOLND	(1<<0)
#define DMA_LIST_DESC_EOLSD	(1<<0)

#undef DAR

struct dma_reg {
	volatile u32 MR;	/* 0x00 - Mode */
	volatile u32 SR;	/* 0x04 - Status */
	u32 pad1;		/* 0x08 - Resv */
	volatile u32 CLNDAR;	/* 0x0c - Current Link Descriptor Addrsess */
	volatile u32 SATR;	/* 0x10 - Source Attributes */
	volatile u32 SAR;	/* 0x14 - Source Address */
	volatile u32 DATR;	/* 0x18 - Destination Attributes */
	volatile u32 DAR;	/* 0x1c - Destination Address */
	volatile u32 BCR0;	/* 0x20 - Byte Count */
	u32 pad2;		/* 0x24 - Resv */
	volatile u32 NLNDAR;	/* 0x28 - Next Link Descriptor Addrsess */
	u32 pad3;		/* 0x2c - Resv */
	u32 pad4;		/* 0x30 - Resv */
	volatile u32 CLSDAR;	/* 0x34 - Current List Descriptor Addrsess */
	u32 pad5;		/* 0x38 - Resv */
	volatile u32 NLSDAR;	/* 0x3c - Next List Descriptor Addrsess */
	volatile u32 SSR;	/* 0x40 - Source Stride */
	volatile u32 DSR;	/* 0x44 - Destination Stride */
};

struct dma_list_desc {
	u32 pad1;		/* 0x00 */
	u32 next_list;		/* 0x04 */
	u32 pad2;		/* 0x08 */
	u32 first_link;		/* 0x0c */
	u32 src_stride;		/* 0x10 */
	u32 dst_stride;		/* 0x14 */
	u32 pad3;		/* 0x18 */
	u32 pad4;		/* 0x1c */
	dma_addr_t paddr;
	struct list_head list;
	struct list_head link;
	int status;
};

struct dma_link_desc {
	u32 src_attr;		/* 0x00 */
	u32 src_addr;		/* 0x04 */
	u32 dst_attr;		/* 0x08 */
	u32 dst_addr;		/* 0x0c */
	u32 pad2;		/* 0x10 */
	u32 next_link;		/* 0x14 */
	u32 count;		/* 0x18 */
	u32 pad4;		/* 0x1c */
	dma_addr_t paddr;
	struct dma_pub pub;
};

struct dma_channel {
	struct pci_dev *pdev;
	int irq;
	int mode;		/* shared or excusive */
	int users;
	struct dma_reg *pdma;	/* CPU address of register set */
	struct tq_struct tq;
	spinlock_t lock;
	struct dma_list_desc *curr_list;	/* */
	struct list_head free_link;	/* list of free link descriptors */
	struct list_head free_list;	/* list of free list descriptors */
	struct list_head pending;
	struct list_head active;
	struct list_head complete;
	void *vpool_link;	/* CPU address of link desc pool */
	dma_addr_t ppool_link;	/* DMA address of link desc pool */
	void *vpool_list;	/* CPU address of list desc pool */
	dma_addr_t ppool_list;	/* DMA address of list desc pool */
};

#define ALIGNED_LINK_DESC_SIZE	((sizeof(struct dma_link_desc) + DMA_ALIGN - 1)  &\
	       			~(DMA_ALIGN - 1))
#define ALIGNED_LIST_DESC_SIZE	((sizeof(struct dma_list_desc) + DMA_ALIGN - 1)  &\
	       			~(DMA_ALIGN - 1))

extern int mpc85xx_dma_init(struct pci_dev *pdev);
extern void mpc85xx_dma_bh(void *data);
extern void mpc85xx_dma_striding_bh(void *data);
extern int mpc85xx_dma_build_desc(struct dma_link_desc *pdesc,
				  struct dma_link_desc *prev,
				  struct dma_channel *ch);

extern int mpc85xx_dma_alloc_list_desc(struct dma_list_desc **list_desc,
				       void *dma_ch);
extern void mpc85xx_dma_append_list_desc(struct list_head *list,
					 struct list_head *head);
extern void __mpc85xx_dma_free_link(struct list_head *head, void *dma_ch);

extern void mpc85xx_dma_append_list(struct list_head *list,
				    struct list_head *head);

#endif				/* __MPC85XX_DMA_PRIV__ */
